# -*- coding: utf-8 -*-
from __future__ import unicode_literals


class Posts(list):
    """
    Represents a list of posts in a thread.
    """

    def __init__(self, *args):
        """
        Initialise a new posts list.

        :param args: The list of posts.
        """

        super(Posts, self).__init__(*args)

    def filter(self, predicate):
        """
        Take a subset of this list of posts.

        :param predicate: The predicate to use to choose which posts make the
                          cut.
        :return: The filtered posts.
        """

        return Posts(filter(predicate, self))

    def map(self, transformation):
        """
        Applies a transformation function to each post, returning a list of this
        function's returned values.

        :param transformation: The transformation function.
        :return: The transformed list of posts.
        """

        return map(transformation, self)

    def foreach(self, function):
        """
        Call a function  for each post.

        :param function: A function taking a post argument. Return values are
        ignored.
        """

        for post in self:
            function(post)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
import contextlib
import re
import json
import six
from six.moves.urllib import request

from chandl.model.posts import Posts
from chandl.model.post import Post


logger = logging.getLogger(__name__)


@six.python_2_unicode_compatible
class Thread:
    """
    Represents a 4Chan thread.
    """

    def __init__(self, board, id_, subject, slug, posts):
        """
        Initialise a new thread instance.

        :param board: The board this thread exists within.
        :param id_: The post id of the first post in this thread.
        :param subject: The thread subject, if set.
        :param slug: The thread's URL fragment.
        :param posts: Posts within this thread.
        """

        self.board = board
        self.id = id_
        self.subject = subject
        self.slug = slug
        self.posts = posts

    @property
    def url(self, secure=True):
        """
        Generate a link where this thread can be viewed.

        :param secure: Whether or not to generate an HTTPS link.
                       Defaults to true.
        :return: The URL.
        """

        protocol = 'https' if secure else 'http'
        return '{0}://boards.4chan.org/{1}/thread/{2}/{3}'.format(
            protocol, self.board, self.id, self.slug)

    @staticmethod
    def parse_json(board, json_):
        """
        Create a thread instance from JSON returned by the 4Chan API.

        :param board: The board that was requested.
        :param json_: The thread's parsed JSON as a dictionary.
        :return: The created thread instance.
        """

        if not json_['posts']:
            raise ValueError('A thread must contain at least one post')

        first = json_['posts'][0]

        return Thread(board,
                      first['no'],
                      first['sub'] if 'sub' in first else None,
                      first['semantic_url'],
                      Posts([Post.parse_json(board, post)
                             for post in json_['posts']]))

    @staticmethod
    def from_url(url):
        """
        Construct a thread instance from its URL.

        :param url: The URL of the thread to retrieve.
        :return: The created thread instance.
        """

        # extract the board and thread ids
        result = re.search('boards\.4chan\.org/([a-z]+)/thread/([0-9]+)', url)
        if not result:
            raise ValueError('Invalid thread URL: {0}'.format(url))

        # download the JSON
        api_url = 'https://a.4cdn.org/{0}/thread/{1}.json'.format(
            result.group(1), result.group(2))
        logger.debug('Retrieving JSON from %s', api_url)
        with contextlib.closing(request.urlopen(api_url)) as response:
            # urllib doesn't decode for us, which confuses json
            text = response.read().decode('utf-8')
            return Thread.parse_json(result.group(1), json.loads(text))

    def __str__(self):
        return 'Thread({0}, {1})'.format(self.id, self.board)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
import os
import sys
import binascii
import base64
import six
from six.moves.urllib import request

from chandl import util


logger = logging.getLogger(__name__)

TYPE_VIDEO = ['webm', 'gif']
TYPE_IMAGE = ['jpg', 'png']


@six.python_2_unicode_compatible
class File:
    """
    Represents a media file attached to a post.
    """

    def __init__(self, id_, board, name, extension, size, width, height, md5):
        """
        Initialise a new file instance.

        :param id_: A unique identifier for the file: it's UNIX timestamp
                    prepended to the number of milliseconds.
        :param board: The board this file was posted in, e.g. 'gif'.
        :param name: The original name of the file.
        :param extension: The file's extension, without leading dot.
        :param size: The size of the file in bytes.
        :param width: The width of the media in pixels.
        :param height: The height of the media in pixels.
        :param md5: The MD5 hash of the file as a 32-character string.
        """

        self.id = id_
        self.board = board
        self.name = name[:30]  # the length of some file names causes problems
        self.extension = extension
        self.size = size
        self.width = width
        self.height = height
        self.md5 = md5

    @property
    def url(self, secure=True):
        """
        Retrieve the URL where this file can be downloaded.

        :param secure: Whether or not to generate an HTTPS link.
                       True by default.
        :return: The URL.
        """

        protocol = 'https' if secure else 'http'
        return '{0}://i.4cdn.org/{1}/{2}.{3}'.format(protocol, self.board,
                                                     self.id, self.extension)

    def save_to(self, directory, name_fmt, verify=True):
        """
        Download and save this file.

        :param directory: The directory to save this file within.
        :param name_fmt: A function taking this instance and returning a file
                         name.
        :param verify: Whether to verify the file's checksum once it is written
                       (default: true).
        :raise IOError: If the file cannot be written, or its checksum does not
                        match the one reported by 4Chan.
        """

        # N.B. urllib doesn't handle unicode well - encode everything before
        #      handing over

        destination = os.path.join(directory, name_fmt(self))\
            .encode(sys.getfilesystemencoding())
        if os.path.isfile(destination):
            logger.info('%s already exists; skipping download', self)
        else:
            logger.info('Downloading %s', self)
            request.urlretrieve(self.url, destination)
        if verify and util.md5_file(destination) != self.md5:
            raise IOError('Verify failed: checksum mismatch'.format(self))

    @staticmethod
    def parse_json(board, json):
        """
        Create a file object from 4Chan's API format.

        :param board: The board this file was uploaded to.
        :param json: The JSON of the post containing this file.
        :return: The created file instance.
        :raises ValueError: If the post does not contain an image.
        """

        def unpack_hash(encoded):
            """
            Unpacks a packed, base64 encoded MD5 checksum.
            :param encoded: The unicode representation of the encoded checksum.
            :return: The unicode representation of the unpacked, 32-character
                     checksum.
            """

            return binascii.hexlify(
                base64.b64decode(encoded.encode('ascii'))).decode('ascii')

        if 'filename' not in json:
            raise ValueError('Post does not contain an image')

        return File(json['tim'], board, json['filename'], json['ext'][1:],
                    json['fsize'], json['w'], json['h'],
                    unpack_hash(json['md5']).rstrip())

    def __str__(self):
        return 'File({0}, {1}.{2}, {3}, {4}x{5})'.format(
            self.id, self.name, self.extension, util.bytes_fmt(self.size),
            self.width, self.height)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
import sys
import hashlib


def bytes_fmt(num, suffix='B'):
    """
    Turn a number of bytes into a more friendly representation, e.g. 2.5MiB.

    :param num: The number of bytes to convert.
    :param suffix: The unit suffix (defaults to 'B').
    :return: The human-readable equivalent of the input size.
    """

    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return '{0:.1f} {1}{2}'.format(num, unit, suffix)
        num /= 1024.0
    return '{0:.1f} {1}{2}'.format(num, 'Yi', suffix)


def decode_cli_arg(arg):
    """
    Turn a bytestring provided by `argparse` into unicode.

    :param arg: The bytestring to decode.
    :return: The argument as a unicode object.
    """
    return arg.decode(sys.getfilesystemencoding())


def make_filename(string):
    """
    Turn a string into something that can be safely used as a file or directory
    name.

    :param string: The string to convert.
    :return: The sanitised string.
    """

    safe = [' ', '.', '_', ':']
    return ''.join(c for c in string if c.isalnum() or c in safe).rstrip()


def md5_file(path):
    """
    Get the MD5 hash of a file.

    :param path: The path of the file.
    :return: The 32-character long lowercase hex representation of the
             checksum.
    """

    hash_ = hashlib.md5()
    with open(path, 'r') as fd:
        for chunk in iter(lambda: fd.read(4096), b''):
            hash_.update(chunk)
    return hash_.hexdigest()


def log_level_from_vebosity(verbosity):
    """
    Get the `logging` module log level from a verbosity.

    :param verbosity: The number of times the `-v` option was specified.
    :return: The corresponding log level.
    """

    if verbosity == 0:
        return logging.WARNING
    if verbosity == 1:
        return logging.INFO
    return logging.DEBUG



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
import multiprocessing
from functools import partial
from threading import Thread
from collections import deque


logger = logging.getLogger(__name__)


class Downloader:
    """
    A basic thread pool implementation to download multiple files
    simultaneously.
    """

    def __init__(self, directory, name_fmt, parallelism=4):
        """
        Initialise a new downloader instance. Instances can be reused.

        :param directory: The directory to save files in.
        :param name_fmt: A function taking a file instance and returning the
                         name to save it under.
        :param parallelism: The maximum number of threads to use to download
                            files per CPU. E.g. parallelism of 4 on a quad core
                            results in 20 threads.
        """

        self._directory = directory
        self._name_fmt = name_fmt
        self._threads = multiprocessing.cpu_count() * parallelism
        self._queue = deque()
        self._pool = []  # invariant: len(self._pool) <= self._threads

    @staticmethod
    def runner(self):
        """
        A single thread's execution.

        :param self: The downloader instance the thread belongs to.
        """

        while True:
            try:
                file_ = self._queue.popleft()
                Downloader.handle(self, file_)
            except IndexError:
                # no items left to process - let function return
                break

    @staticmethod
    def handle(self, file_):
        """
        Downloads a file.

        :param self: The downloader context.
        :param file_: The file to download.
        """

        try:
            file_.save_to(self._directory, self._name_fmt)
        except IOError as e:
            logger.exception('Failed to write %s: %s', file_, e.message)

    def _queue_all(self, files):
        """
        Add all files in an iterable to the download queue.

        :param files: The files to add.
        """

        for file_ in files:
            self._queue.append(file_)

    def download(self, files):
        """
        Download a set of files.

        :param files: An iterable containing the files to download. They will
                      be downloaded in order.
        """

        # populate the queue
        self._queue_all(files)

        # don't launch more threads than files
        threads = min(self._threads, len(self._queue))
        logger.debug('Will use %d threads for downloading', threads)

        # launch threads
        target = partial(Downloader.runner, self)
        for _ in range(threads):
            thread = Thread(target=target)
            thread.start()
            self._pool.append(thread)
        logger.debug('All threads launched')

        # wait for all threads to finish
        for i in range(threads):
            self._pool[i].join()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from datetime import datetime
import six

from chandl.model.file import File


@six.python_2_unicode_compatible
class Post:
    """
    Represents a 4Chan post.
    """

    def __init__(self, board, id_, timestamp, body=None, file_=None):
        """
        Initialise a new post instance.

        :param board: The board this post was submitted to.
        :param id_: The post number.
        :param timestamp: When this post was submitted.
        :param body: The body of this post, if one exists.
        :param file_: This post's file, if one exists.
        """

        self.board = board
        self.id = id_
        self.timestamp = timestamp
        self.body = body
        self.file = file_

    def has_file(self):
        """
        Find whether this post has a file.

        :return: True if it does, false otherwise.
        """

        return self.file is not None

    @staticmethod
    def parse_json(board, json):
        """
        Create a post instance from JSON returned by the 4Chan API.

        :param board: The board that was requested.
        :param json: The post's parsed JSON as a dictionary.
        :return: The created post instance.
        """

        file_ = File.parse_json(board, json) if 'tim' in json else None
        comment = json['com'] if 'com' in json else None
        return Post(board, json['no'], datetime.utcfromtimestamp(json['time']),
                    comment, file_)

    def __str__(self):
        return 'Post({0}, {1}, {2}, {3})'.format(self.id, str(self.timestamp),
                                                 self.body, self.file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function
import sys
import os
import argparse
import logging
from datetime import datetime

from chandl import util
from chandl.downloader import Downloader
from chandl.model.thread import Thread
from chandl.model.post import Post


logger = logging.getLogger(__name__)


def main():
    """
    Chandl entry point.
    """
    parser = argparse.ArgumentParser(prog='chandl',
                                     description='A lightweight tool for '
                                                 'parsing and downloading '
                                                 '4chan threads.')
    parser.add_argument('url',
                        nargs='+',
                        type=util.decode_cli_arg,
                        help='the URL(s) of the thread(s) whose files to '
                             'download')
    parser.add_argument('-c', '--cwd',
                        action='store_true',
                        default=False,
                        help='download to the working directory')
    parser.add_argument('-t', '--threads',
                        type=int,
                        help='the maximum number of download threads to use '
                             'per core',
                        default=5)
    parser.add_argument('-v', '--verbosity',
                        help='increase output verbosity',
                        action='count',
                        default=0)
    args = parser.parse_args()

    # sort out logging output and level
    level = util.log_level_from_vebosity(args.verbosity)
    root = logging.getLogger()
    root.setLevel(level)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(level)
    handler.setFormatter(logging.Formatter('%(levelname)s %(message)s'))
    root.addHandler(handler)

    for url in args.url:
        try:
            thread = Thread.from_url(url)
        except ValueError as e:
            # almost definitely invalid url
            print(e, file=sys.stderr)

        print('Downloading {0}'.format(
            thread.subject if thread.subject else str(thread.id)))
        files = list(thread.posts.filter(Post.has_file).map(
            lambda post: post.file))  # list() as we need to rewind

        if args.cwd:
            directory = os.getcwd()
        else:
            directory = util.make_filename(thread.subject) if thread.subject \
                else str(thread.id)
            if not os.path.exists(directory):
                os.mkdir(directory)

        logger.debug('Downloading to %s', directory)

        dl = Downloader(directory,
                        lambda image: str(image.id) + '.' + image.extension,
                        parallelism=args.threads)
        start = datetime.now()
        dl.download(files)
        finish = datetime.now()

        size = sum([file_.size for file_ in files])
        delta = finish - start
        print('Downloaded {0} in {1:0.3f} seconds ({2}/s)'.format(
            util.bytes_fmt(size),
            delta.total_seconds(),
            util.bytes_fmt(size / delta.total_seconds())))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
import codecs
from setuptools import setup, find_packages


def _read_file(name, encoding='utf-8'):
    """
    Read the contents of a file.

    :param name: The name of the file in the current directory.
    :param encoding: The encoding of the file; defaults to utf-8.
    :return: The contents of the file.
    """
    with codecs.open(name, encoding=encoding) as f:
        return f.read()


setup(
    name='chandl',
    version='0.0.3',
    description='A lightweight tool for parsing and downloading 4chan threads.',
    long_description=_read_file('README.md'),
    license='MIT',
    url='https://github.com/gebn/chandl',
    author='George Brighton',
    author_email='oss@gebn.co.uk',
    packages=find_packages(),
    install_requires=[
        'six>=1.9.0'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Topic :: Utilities'
    ],
    entry_points={
        'console_scripts': [
            'chandl = chandl.__main__:main',
        ]
    }
)
